from django.conf import settings

from telegram import Bot

from django_telegram.bot.constants import SETTINGS_CHAT_ID, SETTINGS_MW, SETTINGS_TOKEN


def send_message(message):
    try:
        bot = Bot(settings.TELEGRAM_BOT[SETTINGS_TOKEN])
        bot.send_message(settings.TELEGRAM_BOT[SETTINGS_MW][SETTINGS_CHAT_ID], message)
        return True
    except Exception:
        #  we do not want this to affect any operations
        return False



def render_as_list(queryset):
    qs_size = len(queryset)
    if qs_size <= 10:
        data_string = '\n'.join(['- {name} ({id}): {status}\n'.format(
            name=i["name"], id=i["id"], status=i["status"],
        ) for i in queryset])
        return f'``` Total: {qs_size}\n{data_string} ```'
    else:
        data_string = '\n'.join(['- {name} ({id}): {status}\n'.format(
            name=i["name"], id=i["id"], status=i["status"],
        ) for i in queryset[0:9]])
        return f'``` Total: {qs_size}\n{data_string}\n- ... and {qs_size - 10} more ...```'



from django.core.management.base import BaseCommand  # pragma: no cover

from django_telegram.bot.bot_runner import BotRunner  # pragma: no cover


class Command(BaseCommand):  # pragma: no cover
    help = "Start Django Telegram bot."

    def handle(self, *args, **options):
        bot_runner = BotRunner()
        bot_runner.handle()



from functools import wraps


def chat_context(func):
    @wraps(func)
    def check_chat_and_exec(self, update, context):
        if self.chat_id == update.message.chat.id:
            res = func(self, update, context)
            return res

    return check_chat_and_exec



from functools import wraps


def log_args(func):
    @wraps(func)
    def log_and_exec(self, update, context):
        self.logger.info(f'Input: {func.__name__}: {update.message.text}')
        res = func(self, update, context)
        self.logger.info(f'Output: {func.__name__}: {res}')
        return res

    return log_and_exec



import warnings

from django.core.exceptions import ImproperlyConfigured
from django.utils.module_loading import import_string

from django_telegram.bot.constants import (
    SETTINGS_CHAT_ID, SETTINGS_COMMANDS_SUFFIX,
    SETTINGS_CONVERSATIONS, SETTINGS_HISTORY_LOOKUP_MODEL_PROPERTY,
    SETTINGS_MW, SETTINGS_MW_CONDITIONS,
    SETTINGS_MW_CONDITIONS_FIELD, SETTINGS_MW_CONDITIONS_FIELD_VALUE,
    SETTINGS_MW_CONDITIONS_FUNC, SETTINGS_MW_CONDITIONS_TYPE,
    SETTINGS_MW_CONDITIONS_VALUE, SETTINGS_MW_MESSAGE,
    SETTINGS_MW_RULES, SETTINGS_MW_TRIGGER_CODES,
    SETTINGS_MW_VIEW, SETTINGS_TOKEN,
)


class TelegramBotConfigurator(object):

    def __init__(self, telegram_settings, django_mw_list):
        self.telegram_settings = telegram_settings
        self.django_mw_list = django_mw_list

    def _check_mw_config_rule_condition(self, condition):
        if SETTINGS_MW_CONDITIONS_TYPE not in condition.keys():
            raise ImproperlyConfigured(
                f'Condition "{SETTINGS_MW_CONDITIONS_TYPE}" key has not been set',
            )

        cond_type = condition[SETTINGS_MW_CONDITIONS_TYPE]
        cond_types = [SETTINGS_MW_CONDITIONS_FUNC, SETTINGS_MW_CONDITIONS_VALUE]
        if cond_type not in cond_types:
            raise ImproperlyConfigured(
                f'Condition "{SETTINGS_MW_CONDITIONS_TYPE}" key must be one of "{cond_types}"',
            )

        if cond_type == SETTINGS_MW_CONDITIONS_FUNC:
            try:
                import_string(condition[SETTINGS_MW_CONDITIONS_FUNC])
            except (KeyError, ImportError):
                raise ImproperlyConfigured(
                    f'Condition "{SETTINGS_MW_CONDITIONS_FUNC}" key must be set and have value. ',
                    'Or specified function could be found.',
                )

        if cond_type == SETTINGS_MW_CONDITIONS_VALUE:
            if SETTINGS_MW_CONDITIONS_FIELD not in condition.keys():
                raise ImproperlyConfigured(
                    f'Condition "{SETTINGS_MW_CONDITIONS_FIELD}" key must be set',
                )
            if not condition[SETTINGS_MW_CONDITIONS_FIELD]:
                raise ImproperlyConfigured(
                    f'Condition "{SETTINGS_MW_CONDITIONS_FIELD}" key is empty',
                )
            if SETTINGS_MW_CONDITIONS_FIELD_VALUE not in condition.keys():
                raise ImproperlyConfigured(
                    f'Condition "{SETTINGS_MW_CONDITIONS_FIELD_VALUE}" key must be set',
                )
            if not condition[SETTINGS_MW_CONDITIONS_FIELD_VALUE]:
                raise ImproperlyConfigured(
                    f'Condition "{SETTINGS_MW_CONDITIONS_FIELD_VALUE}" key is empty',
                )

    def _check_mw_rule(self, config):
        keys = config.keys()
        if SETTINGS_MW_VIEW not in keys or not config[SETTINGS_MW_VIEW]:
            raise ImproperlyConfigured(
                f'"{SETTINGS_MW_VIEW}" key has not been set',
            )

        if SETTINGS_MW_MESSAGE not in keys or not config[SETTINGS_MW_MESSAGE]:
            raise ImproperlyConfigured(
                f'"{SETTINGS_MW_MESSAGE}" key has not been set',
            )

        if SETTINGS_MW_TRIGGER_CODES not in keys or not config[SETTINGS_MW_TRIGGER_CODES]:
            raise ImproperlyConfigured(
                f'"{SETTINGS_MW_TRIGGER_CODES}" key has not been set',
            )

        if not isinstance(config[SETTINGS_MW_TRIGGER_CODES], list):
            raise ImproperlyConfigured(
                f'"{SETTINGS_MW_TRIGGER_CODES}" object must be a list.',
            )

        if any(filter(lambda x: not isinstance(x, int), config[SETTINGS_MW_TRIGGER_CODES])):
            raise ImproperlyConfigured(
                f'"{SETTINGS_MW_TRIGGER_CODES}" contains non-integer values',
            )

        if SETTINGS_MW_CONDITIONS in keys:
            self._check_mw_config_rule_condition(config[SETTINGS_MW_CONDITIONS])

    def _check_mw_settings(self):
        mw_fqdn = 'django_telegram.bot.middleware.TelegramMiddleware'
        if mw_fqdn not in self.django_mw_list:
            return

        settings_keys = self.telegram_settings.keys()
        if SETTINGS_MW not in settings_keys:
            raise ImproperlyConfigured(
                f'"{mw_fqdn}" is enabled, however "{SETTINGS_MW}" key has not been set.',
            )

        if not self.telegram_settings[SETTINGS_MW]:
            raise ImproperlyConfigured(
                f'"{mw_fqdn}" is enabled, however "{SETTINGS_MW}" is empty.',
            )

        if SETTINGS_CHAT_ID not in self.telegram_settings[SETTINGS_MW].keys():
            raise ImproperlyConfigured(f'"{SETTINGS_MW}[{SETTINGS_CHAT_ID}]" key has not been set.')

        if not self.telegram_settings[SETTINGS_MW][SETTINGS_CHAT_ID]:
            raise ImproperlyConfigured(f'"{SETTINGS_MW}[{SETTINGS_CHAT_ID}]" key has no value set.')

        if SETTINGS_MW_RULES not in self.telegram_settings[SETTINGS_MW].keys():
            raise ImproperlyConfigured(
                f'"{SETTINGS_MW}[{SETTINGS_MW_RULES}]" key has not been set.',
            )

        if not isinstance(self.telegram_settings[SETTINGS_MW][SETTINGS_MW_RULES], list):
            raise ImproperlyConfigured(
                f'"{SETTINGS_MW}[{SETTINGS_MW_RULES}]" object must be a list.',
            )

        for setting in self.telegram_settings[SETTINGS_MW][SETTINGS_MW_RULES]:
            index = self.telegram_settings[SETTINGS_MW][SETTINGS_MW_RULES].index(setting)
            try:
                self._check_mw_rule(setting)
            except ImproperlyConfigured as err:
                raise ImproperlyConfigured(
                    f'"{SETTINGS_MW}[{SETTINGS_MW_RULES}]" position "{index}" error: {str(err)}',
                )

    def run_check(self):
        settings_keys = self.telegram_settings.keys()
        if SETTINGS_TOKEN not in settings_keys:
            raise ImproperlyConfigured(
                f'"{SETTINGS_TOKEN}" key has not been set.',
            )
        if not self.telegram_settings[SETTINGS_TOKEN]:
            raise ImproperlyConfigured(
                f'"{SETTINGS_TOKEN}" key has no value set.',
            )
        if SETTINGS_COMMANDS_SUFFIX not in settings_keys:
            raise ImproperlyConfigured(
                f'"{SETTINGS_COMMANDS_SUFFIX}" key has not been set.',
            )
        if SETTINGS_HISTORY_LOOKUP_MODEL_PROPERTY not in settings_keys:
            raise ImproperlyConfigured(
                f'"{SETTINGS_HISTORY_LOOKUP_MODEL_PROPERTY}" key has not been set.',
            )
        if not self.telegram_settings[SETTINGS_HISTORY_LOOKUP_MODEL_PROPERTY]:
            raise ImproperlyConfigured(
                f'"{SETTINGS_HISTORY_LOOKUP_MODEL_PROPERTY}" key has no value set.',
            )
        if SETTINGS_CONVERSATIONS not in settings_keys:
            raise ImproperlyConfigured(
                f'"{SETTINGS_CONVERSATIONS}" key has not been set.',
            )
        if not isinstance(self.telegram_settings[SETTINGS_CONVERSATIONS], list):
            raise ImproperlyConfigured(
                f'"{SETTINGS_CONVERSATIONS}" must be a list of strings.',
            )

        if len(self.telegram_settings[SETTINGS_CONVERSATIONS]) == 0:
            warnings.warn(
                'Conversations list is empty, nothing will be setup for Telegram',
            )

        # middleware settings
        self._check_mw_settings()



import logging

from django.utils.module_loading import import_string

from telegram import ReplyKeyboardRemove
from telegram.ext import Updater

from django.conf import settings

from django_telegram.bot.constants import LOGGER_NAME
from django_telegram.bot.constants import (
    SETTINGS_COMMANDS_SUFFIX, SETTINGS_CONVERSATIONS,
    SETTINGS_HISTORY_LOOKUP_MODEL_PROPERTY, SETTINGS_TOKEN,
)


class BotRunner(object):  # pragma: no cover

    def __init__(self):
        self.token = settings.TELEGRAM_BOT[SETTINGS_TOKEN]
        self.suffix = settings.TELEGRAM_BOT[SETTINGS_COMMANDS_SUFFIX]
        self.model_property = settings.TELEGRAM_BOT[SETTINGS_HISTORY_LOOKUP_MODEL_PROPERTY]
        self.conv_classes = settings.TELEGRAM_BOT[SETTINGS_CONVERSATIONS]

    def get_conversation_handler(self, conv_class):
        try:
            class_def = import_string(conv_class)

            return class_def(
                logger=logging.getLogger(LOGGER_NAME),
                model_datetime_property=self.model_property,
                suffix=self.suffix,
            ).get_conversation_handler()
        except ImportError:
            return None

    @staticmethod
    def error_callback(update, context):
        update.message.reply_text(
            f'*ERR*: ``` {context.error} ```',
            parse_mode='markdown',
            reply_markup=ReplyKeyboardRemove(selective=True),
            reply_to_message_id=update.message.message_id,
            api_kwargs={'chat_id': update.message.chat.id},
        )

    def handle(self, *args, **options):
        updater = Updater(self.token)
        logger = logging.getLogger(LOGGER_NAME)
        for conversation in self.conv_classes:
            logger.info(f'Setting up conversation handler {conversation} ..')
            handler = self.get_conversation_handler(conversation)
            if handler:
                updater.dispatcher.add_handler(handler)
                logger.info(f'    > {conversation}: registered')
            else:
                logger.error(f'    > {conversation}: not registered')
        updater.dispatcher.add_error_handler(self.error_callback)
        logger.info('Setting up error handler ..')
        updater.start_polling()
        logger.info('Connect Reports Bot started!')
        updater.idle()
        logger.info('Connect Reports Bot stopped!')



from django.apps import AppConfig  # pragma: no cover
from django.conf import settings  # pragma: no cover
from django.core.exceptions import ImproperlyConfigured  # pragma: no cover

from django_telegram.configurator import TelegramBotConfigurator  # pragma: no cover


class TelegramBotConfig(AppConfig):  # pragma: no cover
    name = 'django_telegram'
    verbose_name = "Django Telegram Bot"

    def ready(self):
        telegram_settings = getattr(settings, 'TELEGRAM_BOT', None)
        if not telegram_settings:
            raise ImproperlyConfigured(
                "Error obtaining telegram bot settings. ",
                "Please check the documentation on how to configure settings.py ",
                "in your project. TELEGRAM_BOT object is missing.",
            )
        if not isinstance(telegram_settings, dict):
            raise ImproperlyConfigured(
                "TELEGRAM_BOT object must be a dictionary.",
            )
        checker = TelegramBotConfigurator(telegram_settings, settings.MIDDLEWARE)
        checker.run_check()



import operator
import re
from io import StringIO
from abc import ABCMeta
from datetime import timedelta
from functools import reduce
from enum import Enum

import django
from django.db.models import Q, Sum
from django.utils import timezone
from django.core.management import call_command

from telegram import KeyboardButton, ReplyKeyboardMarkup, ReplyKeyboardRemove
from telegram.ext import CommandHandler, ConversationHandler, Filters, MessageHandler

from django_telegram.bot.constants import (
    BTN_CAPTION_BUILD_QUERY, BTN_CAPTION_CUSTOM_MGMT, BTN_CAPTION_USE_SAVED_FILTER,
    COUNT, DAYS, HOURS, NO, SUM, WEEKS, YES,
)
from django_telegram.bot.decorators.chat_context import chat_context
from django_telegram.bot.decorators.log_args import log_args
from django_telegram.bot.errors.saved_filter_not_found import SavedFilterNotFound
from django_telegram.bot.renderers.qs2md import render_as_list


class TelegramConversation(object, metaclass=ABCMeta):
    SUFFIX_REGEXP = r'^[\da-z_]{1,32}$'
    EMPTY_RESULT = 'Nothing found.'

    class STATUS(Enum):
        MODE_SELECTOR = 1
        BUILD = 2
        SAVED = 3
        SAVED_FILTER_SELECT = 11
        BUILD_PERIOD = 4
        BUILD_PERIOD_QUANTITY = 5
        BUILD_FILTERS_YES_NO = 6
        BUILD_FILTERS = 7
        BUILD_AGGREGATE_YES_NO = 8
        BUILD_AGGREGATE = 9
        BUILD_AGGREGATE_SUM_PROPERTY = 10
        CUSTOM_MGMT_COMMAND_SELECT = 12

    def __init__(self, logger, model_datetime_property, suffix=None):
        if suffix and not re.match(self.SUFFIX_REGEXP, suffix):
            raise ValueError(f'Suffix does not match {self.SUFFIX_REGEXP}')

        self.query_context = {}
        self.name = self.__class__.__name__.lower()
        self.logger = logger
        self.model = object
        self.suffix = suffix
        self.model_datetime_property = model_datetime_property
        self._default_query_context()
        self.chat_id = None
        if self.suffix:
            self.entrypoint = f'{self.__class__.__name__.lower()}_{self.suffix}'
            self.fallback = f'cancel_{self.suffix}'
        else:
            self.entrypoint = self.__class__.__name__.lower()
            self.fallback = 'cancel'

    def set_entrypoint_name(self, entrypoint_name):
        self.entrypoint = entrypoint_name

    def set_fallback_name(self, fallback_name):
        self.fallback = fallback_name

    def _default_query_context(self):
        self.query_context = {
            'mode': '',
            'period': {
                'uom': '',
                'quantity': 0,
                'timedelta': None,
            },
            'filters': [],
            'aggregate': {
                'type': '',
                'property': '',
            },
            'saved': '',
            'custom_command': '',
        }

    def _reply(self, update, data, keyboard=None):
        if keyboard:
            reply_keyboard = ReplyKeyboardMarkup(
                keyboard=keyboard,
                resize_keyboard=True,
                selective=True,
            )
        else:
            reply_keyboard = ReplyKeyboardRemove(selective=True)

        if type(data) in [django.db.models.query.QuerySet, list]:
            update.message.reply_text(
                render_as_list(data),
                parse_mode='markdown',
                reply_markup=reply_keyboard,
                reply_to_message_id=update.message.message_id,
                api_kwargs={'chat_id': self.chat_id},
            )
        else:
            update.message.reply_text(
                f'``` {data} ```',
                parse_mode='markdown',
                reply_markup=reply_keyboard,
                reply_to_message_id=update.message.message_id,
                api_kwargs={'chat_id': self.chat_id},
            )

    @property
    def saved_filter_regex(self):
        return f'^({"|".join(self.saved_filters)})$'

    @property
    def saved_filters(self):
        return []

    @property
    def custom_commands_regex(self):
        return f'^({"|".join(self.custom_commands)})$'

    @property
    def custom_commands(self):
        return []

    def set_saved_filter(self, s_filter):
        self.query_context['saved'] = s_filter

    @property
    def saved_filter(self):
        return self.query_context['saved']

    def set_custom_command(self, command):
        self.query_context['custom_command'] = command

    @property
    def custom_command(self):
        return self.query_context['custom_command']

    @property
    def saved_filter_selected(self):
        return self.saved_filter != ''

    @property
    def custom_command_selected(self):
        return self.custom_command != ''

    def set_aggregate_type(self, a_type):
        self.query_context['aggregate']['type'] = a_type

    def set_aggregate_property(self, a_property):
        self.query_context['aggregate']['property'] = a_property

    @property
    def aggregate_type(self):
        return self.query_context['aggregate']['type']

    @property
    def aggregate_property(self):
        return self.query_context['aggregate']['property']

    @property
    def has_aggregate(self):
        return self.query_context['aggregate']['type'] != ''

    def add_query_filter(self, key, value):
        self.query_context['filters'].append({key: value})

    @property
    def query_filters(self):
        return self.query_context['filters']

    def set_query_mode(self, mode):
        self.query_context['mode'] = mode

    @property
    def query_mode(self):
        return self.query_context['mode']

    def set_query_period_uom(self, uom):
        self.query_context['period']['uom'] = uom

    @property
    def query_period_uom(self):
        return self.query_context['period']['uom']

    @property
    def query_period_delta(self):
        return timedelta(**{self.query_period_uom: self.query_period_quantity})

    def set_query_period_quantity(self, quantity):
        self.query_context['period']['quantity'] = quantity

    @property
    def query_period_quantity(self):
        return int(self.query_context['period']['quantity'])

    @property
    def has_query_filters(self):
        return len(self.query_context['filters']) > 0

    def set_chat_id(self, user_id):
        self.chat_id = user_id

    def _end_conversation(self, update):
        self._reply(update, 'End of conversation')
        self.set_chat_id(None)
        return ConversationHandler.END

    def execute_custom_command(self, update, command):
        if command not in self.custom_commands:
            self._reply(update, 'Invalid command')
            return False

        out = StringIO()
        try:
            call_command(command, stderr=out, stdout=out)
            self._reply(update, f'Result {command}:\n{out.getvalue()}')
            return True
        except Exception as e:
            self._reply(update, f'Error {command}:\n{e}')
            return False

    @log_args
    def show_mode_select(self, update, context):
        self.set_chat_id(update.message.chat.id)
        self._default_query_context()
        reply_keyboard = [
            [KeyboardButton(text=BTN_CAPTION_BUILD_QUERY)],
            [KeyboardButton(text=BTN_CAPTION_USE_SAVED_FILTER)],
            [KeyboardButton(text=BTN_CAPTION_CUSTOM_MGMT)],
        ]
        self._reply(update, 'How do you want to proceed', reply_keyboard)
        return self.STATUS.MODE_SELECTOR

    def show_list_of_commands(self, update, context):
        self.set_chat_id(update.message.chat.id)
        classes = '\n - '.join([
            f'{cls.__name__.lower()}_{self.suffix}'
            for cls in TelegramConversation.__subclasses__()
        ])
        self._reply(update, f'Available commands:\n - {classes}')
        return ConversationHandler.END

    @log_args
    @chat_context
    def get_mode_show_mode_options(self, update, context):
        self.set_query_mode(update.message.text)
        if self.query_mode == BTN_CAPTION_BUILD_QUERY:
            reply_keyboard = [
                [
                    KeyboardButton(text=DAYS),
                    KeyboardButton(text=WEEKS),
                    KeyboardButton(text=HOURS),
                ],
            ]
            self._reply(update, 'Please select the period unit', reply_keyboard)
            return self.STATUS.BUILD_PERIOD

        if self.query_mode == BTN_CAPTION_USE_SAVED_FILTER:
            if not len(self.saved_filters):
                self._reply(update, 'No saved filters found for this command')
                return ConversationHandler.END
            else:
                reply_keyboard = [
                    [
                        KeyboardButton(text=i)
                        for i in self.saved_filters
                    ],
                ]
                self._reply(update, 'Please select saved filter', reply_keyboard)
                return self.STATUS.SAVED_FILTER_SELECT

        if self.query_mode == BTN_CAPTION_CUSTOM_MGMT:
            if not len(self.custom_commands):
                self._reply(update, 'No custom commands found for this command')
                return ConversationHandler.END
            else:
                reply_keyboard = [
                    [
                        KeyboardButton(text=i)
                        for i in self.custom_commands
                    ],
                ]
                self._reply(update, 'Please select command', reply_keyboard)
                return self.STATUS.CUSTOM_MGMT_COMMAND_SELECT

    @log_args
    @chat_context
    def get_custom_command_and_execute(self, update, context):
        self.set_custom_command(update.message.text)
        self.run_query(update)
        return self._end_conversation(update)

    @log_args
    @chat_context
    def get_period_uom_show_quantity(self, update, context):
        self.set_query_period_uom(update.message.text)
        self._reply(update, 'Please enter the period quantity')
        return self.STATUS.BUILD_PERIOD_QUANTITY

    @log_args
    @chat_context
    def get_quantity_show_yes_no_filters(self, update, context):
        self.set_query_period_quantity(update.message.text)
        reply_keyboard = [
            [KeyboardButton(text=YES)],
            [KeyboardButton(text=NO)],
        ]
        self._reply(update, 'Do you want to specify model filters?', reply_keyboard)
        return self.STATUS.BUILD_FILTERS_YES_NO

    @log_args
    @chat_context
    def get_yes_no_filters_and_proceed(self, update, context):
        if update.message.text == YES:
            self._reply(update, 'Provide model filters')
            return self.STATUS.BUILD_FILTERS
        else:
            reply_keyboard = [
                [KeyboardButton(text=YES)],
                [KeyboardButton(text=NO)],
            ]
            self._reply(update, 'Do you want to specify model aggregate?', reply_keyboard)
            return self.STATUS.BUILD_AGGREGATE_YES_NO

    @log_args
    @chat_context
    def get_yes_no_aggregate_and_proceed(self, update, context):
        if update.message.text == YES:
            reply_keyboard = [
                [
                    KeyboardButton(text=COUNT),
                    KeyboardButton(text=SUM),
                ],
            ]
            self._reply(update, 'Please select the aggregate', reply_keyboard)
            return self.STATUS.BUILD_AGGREGATE

        else:
            self.run_query(update)
            return self._end_conversation(update)

    @log_args
    @chat_context
    def get_filters_and_proceed(self, update, context):
        filters = update.message.text
        for model_filter in filters.split(','):
            field = model_filter.split('=')[0]
            field_value = model_filter.split('=')[1]
            self.add_query_filter(field, field_value)

        reply_keyboard = [
            [KeyboardButton(text=YES)],
            [KeyboardButton(text=NO)],
        ]
        self._reply(update, 'Do you want to specify model aggregate?', reply_keyboard)
        return self.STATUS.BUILD_AGGREGATE_YES_NO

    @log_args
    @chat_context
    def get_aggregate_property_and_proceed(self, update, context):
        self.set_aggregate_property(update.message.text)
        self.run_query(update)
        return self._end_conversation(update)

    @log_args
    @chat_context
    def get_saved_filter_and_proceed(self, update, context):
        self.set_saved_filter(update.message.text)
        self.run_query(update)
        return self._end_conversation(update)

    @log_args
    @chat_context
    def get_aggregate_and_proceed(self, update, context):
        self.set_aggregate_type(update.message.text)
        if self.aggregate_type == COUNT:
            self.run_query(update)
            return ConversationHandler.END

        if self.aggregate_type == SUM:
            self._reply(update, 'Provide property to aggregate')
            return self.STATUS.BUILD_AGGREGATE_SUM_PROPERTY

        return self._end_conversation(update)

    def _get_initial_queryset(self):
        return self.model.objects.all()

    def run_query(self, update):
        if self.saved_filter_selected:
            self.logger.info(f'Saved filter {self.model.__name__} : {self.saved_filter}')
            call_method = getattr(self, f'get_{self.saved_filter}', None)
            if not call_method:
                raise SavedFilterNotFound(f'{self.saved_filter} not found')
            call_method(update)
            return

        if self.custom_command_selected:
            self.logger.info(f'Custom command {self.model.__name__} : {self.custom_command}')
            self.execute_custom_command(update, command=self.custom_command)
            return

        self.logger.info(f'Lookup {self.query_context}')

        history_lookup = {
            f'{self.model_datetime_property}__gt': timezone.now() - self.query_period_delta,
        }
        queryset = self._get_initial_queryset().filter(**history_lookup)

        if self.has_query_filters:
            queryset = queryset.filter(
                reduce(operator.and_, (Q(**d) for d in self.query_filters)),
            )

        queryset = queryset.values('id', 'name', 'status')

        self.logger.info(f'Resulting queryset: {len(queryset)}')

        if self.has_aggregate:
            if self.aggregate_type == COUNT:
                self._reply(update, len(queryset))
                return

            if self.aggregate_type == SUM:
                queryset = queryset.aggregate(Sum(self.aggregate_property))
                self._reply(update, queryset[f'{self.aggregate_property}__sum'])
                return

        if queryset:
            self._reply(update, list(queryset))
        else:
            self._reply(update, self.EMPTY_RESULT)

    @log_args
    @chat_context
    def cancel(self, update, context):
        return self._end_conversation(update)

    def get_conversation_handler(self):
        modes = "|".join([
            BTN_CAPTION_BUILD_QUERY,
            BTN_CAPTION_USE_SAVED_FILTER,
            BTN_CAPTION_CUSTOM_MGMT,
        ])
        mode_sel_re = f'^({modes})$'

        return ConversationHandler(
            entry_points=[
                CommandHandler(self.entrypoint, self.show_mode_select),
                CommandHandler('commands', self.show_list_of_commands),
            ],
            states={
                self.STATUS.MODE_SELECTOR: [
                    MessageHandler(Filters.regex(mode_sel_re), self.get_mode_show_mode_options),
                ],
                self.STATUS.CUSTOM_MGMT_COMMAND_SELECT: [
                    MessageHandler(Filters.regex(
                        self.custom_commands_regex,
                    ), self.get_custom_command_and_execute),
                ],
                self.STATUS.BUILD_PERIOD: [
                    MessageHandler(Filters.regex(
                        f'^({DAYS}|{WEEKS}|{HOURS})$',
                    ), self.get_period_uom_show_quantity),
                ],
                self.STATUS.BUILD_PERIOD_QUANTITY: [
                    MessageHandler(
                        Filters.regex('^(\\d)+$'),
                        self.get_quantity_show_yes_no_filters,
                    ),
                ],
                self.STATUS.BUILD_FILTERS_YES_NO: [
                    MessageHandler(
                        Filters.regex(f'^({YES}|{NO})$'),
                        self.get_yes_no_filters_and_proceed,
                    ),
                ],
                self.STATUS.BUILD_AGGREGATE_YES_NO: [
                    MessageHandler(
                        Filters.regex(f'^({YES}|{NO})$'),
                        self.get_yes_no_aggregate_and_proceed,
                    ),
                ],
                self.STATUS.BUILD_FILTERS: [
                    MessageHandler(Filters.text, self.get_filters_and_proceed),
                ],
                self.STATUS.BUILD_AGGREGATE: [
                    MessageHandler(
                        Filters.regex(f'^({COUNT}|{SUM})$'),
                        self.get_aggregate_and_proceed,
                    ),
                ],
                self.STATUS.BUILD_AGGREGATE_SUM_PROPERTY: [
                    MessageHandler(Filters.text, self.get_aggregate_property_and_proceed),
                ],
                self.STATUS.SAVED_FILTER_SELECT: [
                    MessageHandler(
                        Filters.regex(self.saved_filter_regex),
                        self.get_saved_filter_and_proceed,
                    ),
                ],

            },
            fallbacks=[CommandHandler(self.fallback, self.cancel)],
        )



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['django_telegram',
 'django_telegram.bot',
 'django_telegram.bot.decorators',
 'django_telegram.bot.errors',
 'django_telegram.bot.renderers',
 'django_telegram.management',
 'django_telegram.management.commands']

package_data = \
{'': ['*']}

install_requires = \
['django>=2.2,<2.3', 'python-telegram-bot>=13.3,<14.0']

setup_kwargs = {
    'name': 'telegram-django',
    'version': '1.0.10',
    'description': 'Django Telegram Bot',
    'long_description': "# Django Telegram Bot\n\n![pyversions](https://img.shields.io/pypi/pyversions/telegram-django.svg) [![PyPi Status](https://img.shields.io/pypi/v/telegram-django.svg)](https://pypi.org/project/telegram-django/) [![Build Django Telegram Bot](https://github.com/cloudblue/telegram-django/actions/workflows/build.yml/badge.svg)](https://github.com/cloudblue/telegram-django/actions/workflows/build.yml) [![Quality Gate Status](https://sonarcloud.io/api/project_badges/measure?project=telegram-django&metric=alert_status)](https://sonarcloud.io/dashboard?id=telegram-django) [![Coverage](https://sonarcloud.io/api/project_badges/measure?project=telegram-django&metric=coverage)](https://sonarcloud.io/dashboard?id=telegram-django) [![Maintainability Rating](https://sonarcloud.io/api/project_badges/measure?project=telegram-django&metric=sqale_rating)](https://sonarcloud.io/dashboard?id=telegram-django)\n\n## Introduction\n\n`Django Telegram Bot` is the base class which helps to build custom commands for the django models. It allows getting historical data from django models, as well as develop custom pre-defined filters and execute custom django management commands (i.e. those which are executed through  `python manage.py $command`)\n\n## Install\n\n`Django Telegram Bot` requires python 3.8 or later and has the following dependencies:\n\n* python-telegram-bot >=13.3\n* django>=2.2.20\n\n`Django Telegram Bot` can be installed from [pypi.org](https://pypi.org/project/telegram-django/) using pip:\n\n```\n$ pip install telegram-django\n```\n\n## Running The Bot\n### Define your command class extending the base\n\nFor example:\n\n``` \n\nfrom myapp.models import MyAppModel\n\nfrom telegram_bot.telegram_conversation import TelegramConversation\n\n\nclass MyAppConversation(TelegramConversation):\n    def __init__(self, logger, model_datetime_property, suffix):\n        super().__init__(logger, model_datetime_property, suffix)\n        self.model = MyAppModel\n\n    @property\n    def custom_commands(self):\n        return [\n            'custom_command1',\n        ]\n        \n    @property\n    def saved_filters(self):\n        return [\n            'count',\n        ]\n\n    def get_count(self, update):\n        amount = self._get_initial_queryset().count()\n        self._reply(update, amount)\n        self._default_query_context()\n\n\n```\n\nThe method ```custom_commands``` must return a list of defined django commands which can be executed by this conversation handler. These commands are standard django commands which are normally executed via ```python manage.py $command```.\nThe method ```saved_filters``` must return a list of defined custom filters. The filter's body must be implemented in the same class using the convention ```get_$filter_name```, like in the example above: for ```count``` filter the ```get_count``` method is implemented.\n\nAdd the following sections to your ```settings.py```:\n\nDefine application in ```INSTALLED_APPS```\n```\n    INSTALLED_APPS = [\n        ...\n        'django_telegram',\n        ...\n    ]\n```\nAdd section ```TELEGRAM_BOT``` for bot configuration\n```\nTELEGRAM_BOT = {\n    'CONVERSATIONS': [\n        'myapp.package1.package2.MyAppConversation',\n    ],\n    'TOKEN': '',\n    'COMMANDS_SUFFIX': None,\n    'HISTORY_LOOKUP_MODEL_PROPERTY': ''\n}\n```\nAdd section ```django_telegram_bot``` to logger configuration\n```\nLOGGING = {\n...\n    'loggers': {\n...\n        'django_telegram_bot': {\n            'handlers': ['console'],\n            'level': 'INFO',\n        },\n...\n    },\n...\n}\n```\n\nSettings description:\n\n| Variable      | Description  |\n| ------------- |:-------------|\n|`TOKEN`|Telegram Token for the bot. Please refer to https://core.telegram.org/bots on how to create a bot.|\n|`CONVERSATIONS`|List of FQDNs for classes which implement and provide conversation instances|\n|`HISTORY_LOOKUP_MODEL_PROPERTY`|Property of the django model of DateTime type which is used to do history lookups|\n|`COMMANDS_SUFFIX`|In case of having multiple instances of the bot (with the same commands) we want to add some suffix to the commands, so that only specific bot is getting the command, so command becomes `myappconversation_${SUFFIX}`. If there is no need to have multiple instances of the same bot in the chat -- just leave this as ```None```. |\n\n### Running The Bot\n\n`python manage.py start_bot`\n\n## Middleware\nThe library also provides a way to analyse **responses of type application/json** and based on defined rules send pre-defined messages.\nTo enable middleware add the following line into your ```settings.MIDDLEWARE```\n```\nMIDDLEWARE = [\n...\n    'django_telegram.bot.middleware.TelegramMiddleware',\n]\n```\n\nExtend previously defined configuration in ```settings.py``` with the following\n```\nTELEGRAM_BOT = {\n    'CONVERSATIONS': [\n        'myapp.package1.package2.MyAppConversation',\n    ],\n    'TOKEN': '',\n    'COMMANDS_SUFFIX': None,\n    'HISTORY_LOOKUP_MODEL_PROPERTY': '',\n    'MIDDLEWARE': {\n        'CHAT_ID': -12356789,\n        'RULES': [{\n            'view': 'view',\n            'trigger_codes': [204, 200],\n            'conditions': {\n                'type': 'value',\n                'field': 'field',\n                'field_value': 'value',\n            },\n            'message': 'Message when this condition happened!',\n        }],\n    }\n}\n```\nSettings description:\n\n| Variable      | Description  |\n| ------------- |:-------------|\n|`CHAT_ID`|Telegram chat id to where the bot must send messages. Typically an integer like ```-12345677898```|\n|`RULES`|List of rules objects which configure a case when message must be sent|\n|`RULES[i].view`|View id in resolver definition, i.e. ```/v1/reports/fail/123``` URI would be a ```reports-fail```|\n|`RULES[i].trigger_codes`|List of HTTP codes in response where this rule needs to be triggered|\n|`RULES[i].conditions`|Optional. Additional checks which need to be done before sending the message|\n|`RULES[i].conditions.type`|Type of condition, can be either 'function' (when validation is done by user defined function) or 'value' (when validation is done by simple field/value comparison of response JSON).|\n|`RULES[i].conditions.function`|Required if `RULES[i].conditions.type` is `function` otherwise ignored. User defined function which receives as input response JSON as dict and must return either `True` or `False`|\n|`RULES[i].conditions.field`|Required if `RULES[i].conditions.type` is `value` otherwise ignored. Field to look up in response JSON|\n|`RULES[i].conditions.field_value`|Required if `RULES[i].conditions.type` is `value` otherwise ignored. Expected value to look up in response JSON|\n|`RULES[i].message`|Message which needs to be sent to Telegram in case all conditions match|\n\n\n## Testing\n\n* Create virtualenv\n* Install project dependencies\n```commandline\npython -m pip install --upgrade pip\npip install poetry\npoetry update\n```\n* Run tests\n```commandline\npoetry run pytest\n```\n\n\n## License\n\n``Django Telegram Bot`` is released under the [Apache License Version 2.0](https://www.apache.org/licenses/LICENSE-2.0).",
    'author': 'CloudBlue LLC',
    'author_email': None,
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://connect.cloudblue.com',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



BTN_CAPTION_BUILD_QUERY = 'Build Query'
BTN_CAPTION_USE_SAVED_FILTER = 'Use Saved Filter'
BTN_CAPTION_CUSTOM_MGMT = 'Custom Management Command'
DAYS = 'days'
WEEKS = 'weeks'
HOURS = 'hours'
YES = 'Yes'
NO = 'No'
COUNT = 'count'
SUM = 'sum'
SETTINGS_TOKEN = 'TOKEN'
SETTINGS_COMMANDS_SUFFIX = 'COMMANDS_SUFFIX'
SETTINGS_HISTORY_LOOKUP_MODEL_PROPERTY = 'HISTORY_LOOKUP_MODEL_PROPERTY'
SETTINGS_CONVERSATIONS = 'CONVERSATIONS'
SETTINGS_CHAT_ID = 'CHAT_ID'
SETTINGS_MW = 'MIDDLEWARE'
SETTINGS_MW_VIEW = 'view'
SETTINGS_MW_RULES = 'RULES'
SETTINGS_MW_TRIGGER_CODES = 'trigger_codes'
SETTINGS_MW_CONDITIONS = 'conditions'
SETTINGS_MW_CONDITIONS_TYPE = 'type'
SETTINGS_MW_CONDITIONS_FUNC = 'function'
SETTINGS_MW_CONDITIONS_VALUE = 'value'
SETTINGS_MW_CONDITIONS_FIELD = 'field'
SETTINGS_MW_MESSAGE = 'message'
SETTINGS_MW_CONDITIONS_FIELD_VALUE = 'field_value'
LOGGER_NAME = 'django_telegram_bot'



import json
import logging

from django.conf import settings
from django.utils.deprecation import MiddlewareMixin
from django.utils.module_loading import import_string

from rest_framework import status

from django_telegram.bot.commands import send_message
from django_telegram.bot.constants import (
    LOGGER_NAME, SETTINGS_MW, SETTINGS_MW_CONDITIONS,
    SETTINGS_MW_CONDITIONS_FIELD, SETTINGS_MW_CONDITIONS_FIELD_VALUE,
    SETTINGS_MW_CONDITIONS_FUNC, SETTINGS_MW_CONDITIONS_TYPE,
    SETTINGS_MW_CONDITIONS_VALUE, SETTINGS_MW_MESSAGE, SETTINGS_MW_RULES,
    SETTINGS_MW_TRIGGER_CODES, SETTINGS_MW_VIEW,
)


class TelegramMiddleware(MiddlewareMixin):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.configs = settings.TELEGRAM_BOT[SETTINGS_MW]

    def get_field_value(self, model, field):
        field_parts = field.split('.')
        data = model
        for part in field_parts:
            data = data.get(part, None)
        return data

    def matches_config(self, current_config, response):
        if int(response.status_code) not in current_config[SETTINGS_MW_TRIGGER_CODES]:
            return False

        if SETTINGS_MW_CONDITIONS not in current_config.keys():
            return True

        conditions = current_config[SETTINGS_MW_CONDITIONS]
        cond_type = conditions[SETTINGS_MW_CONDITIONS_TYPE]
        if cond_type == SETTINGS_MW_CONDITIONS_VALUE:
            cond_field = conditions[SETTINGS_MW_CONDITIONS_FIELD]
            cond_value = conditions[SETTINGS_MW_CONDITIONS_FIELD_VALUE]
            field_value = self.get_field_value(json.loads(response.content), cond_field)
            if field_value == cond_value:
                return True

        if cond_type == SETTINGS_MW_CONDITIONS_FUNC:
            user_func = conditions[SETTINGS_MW_CONDITIONS_FUNC]
            user_func_def = import_string(user_func)
            try:
                return user_func_def(json.loads(response.content))
            except Exception:
                return False

    def process_response(self, request, response):
        if response.status_code == status.HTTP_204_NO_CONTENT:
            return response

        if response['content-type'].lower() != "application/json":
            return response

        view_name = request.resolver_match.view_name
        config_exists = any(
            filter(lambda x: x[SETTINGS_MW_VIEW] == view_name, self.configs[SETTINGS_MW_RULES]),
        )
        if config_exists:
            current_config = list(
                filter(lambda x: x[SETTINGS_MW_VIEW] == view_name, self.configs[SETTINGS_MW_RULES]),
            )[0]
            try:
                if self.matches_config(current_config, response):
                    send_message(
                        f'{request.resolver_match.view_name} with pk '
                        f'{request.resolver_match.kwargs.get("pk", None)} '
                        f'has ended with {response.status_code} '
                        f'and sends message: {current_config[SETTINGS_MW_MESSAGE]}',
                    )
            except Exception as e:
                #  we do not want this to affect any operations
                logger = logging.getLogger(LOGGER_NAME)
                logger.error(
                    f'TelegramMiddleware rule {current_config} finished with error: {str(e)}',
                )
        return response
